 
#ifndef __CNF_H
#define __CNF_H

#include <iostream>
#include <string>
#include <vector>
#include <map>

namespace SatSolver { class SatSolver; }

namespace Cnf
{
	class CnfVariable;
	class CnfLiteral;
	class CnfClause;
	class CnfFormula;

	namespace CnfVariableValue
	{
		enum CnfVariableValue
		{
			True,
			False,
			Undefined,
		};
	}

	class CnfVariable
	{
		friend class CnfLiteral;
		friend class CnfFormula;
		friend class SatSolver::SatSolver;
		private:
			CnfVariableValue::CnfVariableValue value;
			std::string name;
			int id;

			std::vector<CnfLiteral*> positives;
			std::vector<CnfLiteral*> negatives;
			int numUnsolvedPositives;
			int numUnsolvedNegatives;

			void RegisterLiteral(bool positive, CnfLiteral* literal);
			CnfVariable(int id, const std::string& name);
			void onClauseSatisfactionChanged(CnfClause* clause, bool positiveLiteral, bool newSatisfaction);
		public:
			~CnfVariable();
			void Write(std::ostream& os, bool humanreadable=false);
			void SetValue(CnfVariableValue::CnfVariableValue newvalue);
			void PrintAssignment(std::ostream& os);
	};

	class CnfLiteral
	{
		friend class CnfVariable;
		friend class CnfClause;
		friend class SatSolver::SatSolver;
		private:
			bool positive;
			CnfVariable* variable;
			CnfClause* clause;
			void onVariableValueChanged(CnfVariableValue::CnfVariableValue oldvalue, CnfVariableValue::CnfVariableValue newvalue);
			void onClauseSatisfactionChanged(CnfClause* clause, bool newSatisfaction);
		public:
			CnfLiteral(bool positive, CnfVariable* variable);
			~CnfLiteral();
			void Write(std::ostream& os, bool humanreadable=false);
	};

	class CnfClause
	{
		friend class CnfLiteral;
		friend class CnfFormula;
		friend class SatSolver::SatSolver;
		private:
			CnfFormula* formula;
			int numSatisfiedLiterals;
			int numUnassignedLiterals;
			std::vector<CnfLiteral*> literals;
			void onVariableValueChanged(CnfVariable* variable, CnfVariableValue::CnfVariableValue oldvalue, CnfVariableValue::CnfVariableValue newvalue);
			void onLiteralSatisfactionChanged(CnfLiteral* literal, bool satisfaction);
			CnfClause(CnfFormula* parent);
		public:
			~CnfClause();
			CnfLiteral* addLiteral(bool positive, CnfVariable* variable);
			void Write(std::ostream& os, bool humanreadable=false);
	};

	class CnfFormula
	{
		friend class CnfClause;
		friend class SatSolver::SatSolver;
		private:
			std::vector<CnfClause*> clauses;
			std::vector<CnfVariable*> variables;
			std::map<std::string, CnfVariable*> variableIndex;
			int numUnsolvedClauses;
			void onClauseSatisfactionChanged(CnfClause* clause, bool newSatisfaction);
		public:
			CnfFormula();
			~CnfFormula();
			CnfVariable* addVariable(std::string name = "");
			CnfClause* addClause(bool p1=false, CnfVariable* v1=nullptr, bool p2=false, CnfVariable* v2=nullptr, bool p3=false, CnfVariable* v3=nullptr);
			void Write(std::ostream& os, bool humanreadable=false);
			void PrintAssignment(std::ostream& os);
	};
}

#endif
