

#include <iostream>
#include <fstream>
#include <string>
#include <map>
#ifdef TARGET_PLATFORM_WASM
    #include <sstream>
    #include <emscripten.h>
#endif

#include "satlexer.h"
#include "parser.hpp"
#include "sat.h"
#include "cnf.h"
#include "satsolver.h"




#ifdef TARGET_PLATFORM_NATIVE
int main(int argc, char** argv)
{
	if(argc < 2)
	{
		std::cerr << argv[0] << " needs an input file as parameter" << std::endl;
		return 1;
	}
	std::ifstream* is = new std::ifstream(argv[1]);
    std::ostream* os = &std::cout;
#endif

#ifdef TARGET_PLATFORM_WASM
extern "C" int SolveSatFormulas(char* message) {
	std::stringstream sis, sos;
	std::istream* is = &sis;
	std::stringstream* os = &sos;
	sis << message;
#endif

	std::vector<Sat::SatFormula*> formulas;
	SatLexer lexer(*is);
	yy::SatParser parser(lexer, &formulas);
	parser.parse();


	if(formulas.size() > 0)
	{
		Cnf::CnfFormula* cnf = formulas[0]->ToCnf();
		for(int i = 1; i < formulas.size(); i++)
		{
			std::vector<Sat::SatFormula*> conjunctions = formulas[i]->CollectConjunctions();
			for(Sat::SatFormula* conjunction : conjunctions)
			{
				Cnf::CnfClause* clause = cnf->addClause();
				
				std::vector<Sat::SatFormula*> disjunctions = conjunction->CollectDisjunctions();
				for(Sat::SatFormula* disjunction : disjunctions)
				{
					bool positive = true;
					Sat::SatFormula* literal = disjunction;
					if(Sat::SatNot* satnot = dynamic_cast<Sat::SatNot*>(literal))
					{
						literal = satnot->f;
						positive = false;
					}
					
					clause->addLiteral(positive, literal->ToCnf(cnf));
				}
			}
		}
		std::cerr << "Conjunctive Normal Form:" << std::endl;
		cnf->Write(std::cerr, true);

		bool solvable = SatSolver::SatSolver::SolveCnf(cnf);
		if(solvable)
			cnf->PrintAssignment(*os);
		else
			(*os) << "Unsolvable" << std::endl;

		delete cnf;
	}
	for(auto f : formulas)
		delete f;

#ifdef TARGET_PLATFORM_WASM
	delete is;
	EM_ASM_ARGS({
		var msg = Pointer_stringify($0);
		ProcessResult(msg);
	}, sos.str().c_str());
#endif

	return 0;
}

